#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.zmi.widget module

This module defines custom widgets which are used to handle thesaurus terms
selection.
"""

import json
import re

from pyramid.decorator import reify
from zope.component import queryMultiAdapter
from zope.interface import directlyProvides, implementer_only
from zope.schema.fieldproperty import FieldProperty
from zope.schema.interfaces import ITitledTokenizedTerm
from zope.schema.vocabulary import SimpleTerm

from pyams_form.browser.select import SelectWidget
from pyams_form.converter import SequenceDataConverter
from pyams_form.interfaces import IDataConverter
from pyams_form.interfaces.widget import IFieldWidget
from pyams_form.widget import FieldWidget
from pyams_layer.interfaces import IFormLayer
from pyams_thesaurus.interfaces.term import IThesaurusTerm, STATUS_ARCHIVED
from pyams_thesaurus.interfaces.thesaurus import IThesaurus
from pyams_thesaurus.schema import IThesaurusTermField, IThesaurusTermsListField
from pyams_thesaurus.zmi.interfaces import IThesaurusTermWidget, IThesaurusTermsListWidget, \
    IThesaurusTermsTreeWidget
from pyams_utils.adapter import adapter_config
from pyams_utils.registry import query_utility
from pyams_utils.traversing import get_parent


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _  # pylint: disable=ungrouped-imports
from pyams_utils.zodb import volatile_property


class SimpleThesaurusTerm(SimpleTerm):
    """Simple thesaurus term"""

    def __init__(self, value, token=None, title=None):  # pylint: disable=super-init-not-called
        self.value = value
        self.token = token or value
        self.title = title
        if title is not None:
            directlyProvides(self, ITitledTokenizedTerm)


def thesaurus_widget_term_factory(term, widget):
    """Thesaurus term factory"""
    if not IThesaurusTerm.providedBy(term):
        thesaurus = widget.thesaurus
        if thesaurus is not None:
            term = thesaurus.terms.get(term)
    if IThesaurusTerm.providedBy(term):
        return SimpleThesaurusTerm(term.label, title=term.label)
    return None


class ThesaurusWidgetMixin:  # pylint: disable=no-member
    """Thesaurus base widget mixin"""

    def term_factory(self, value):
        """Selected term factory"""
        return thesaurus_widget_term_factory(value, self)

    @volatile_property
    def thesaurus(self):
        """Thesaurus utility getter"""
        content = self.form.get_content()
        if IThesaurusTerm.providedBy(content):
            content = get_parent(content, IThesaurus)
        if IThesaurus.providedBy(content):
            thesaurus = content
        else:
            thesaurus = queryMultiAdapter((content, self.request, self.form),
                                          IThesaurus, name=self.field.__name__)
            if thesaurus is None:
                name = self.thesaurus_name
                if name:
                    thesaurus = query_utility(IThesaurus, name=name)
        return thesaurus

    @property
    def thesaurus_name(self):
        """Thesaurus name getter"""
        return self._thesaurus_name

    @thesaurus_name.setter
    def thesaurus_name(self, value):
        """Thesaurus name setter"""
        if value != self._thesaurus_name:
            del self.thesaurus
            self._thesaurus_name = value
            if self.terms is not None:
                self.terms = None
                self.update_terms()

    def is_selected(self, term):
        """Check for term selection"""
        return term.token in self.value

    @property
    def ajax_params(self):
        """AJAX request params getter"""
        return json.dumps({
            'thesaurus_name': self.thesaurus_name,
            'extract_name': self.extract_name
        })


#
# Term widget
#

SYNONYM = re.compile(r'(.*) \[ .* \]')


@adapter_config(required=(IThesaurusTermField, IThesaurusTermWidget),
                provides=IDataConverter)
class ThesaurusTermDataConverter(SequenceDataConverter):
    """Thesaurus term data converter"""

    def to_widget_value(self, value):
        # Widget expects term label or caption
        if isinstance(value, (set, list, tuple)):
            try:
                value = next(iter(value))
            except StopIteration:
                value = None
        if IThesaurusTerm.providedBy(value):
            value = value.label
        return super().to_widget_value(value)

    def to_field_value(self, value):
        # Field expects thesaurus term instance
        if not value:
            return self.field.missing_value
        thesaurus = self.widget.thesaurus
        if thesaurus is None:
            return None
        if isinstance(value, (set, list, tuple)):
            try:
                value = next(iter(value))
            except StopIteration:
                value = None
        if value == self.widget.no_value_token:
            return None
        match = SYNONYM.match(value)
        if match:
            value = match.groups()[0]
        return thesaurus.terms.get(value)


@implementer_only(IThesaurusTermWidget)
class ThesaurusTermWidget(ThesaurusWidgetMixin, SelectWidget):
    """Thesaurus term widget"""

    ajax_url = '/api/thesaurus/terms'
    placeholder = _("No selected term")

    _thesaurus_name = FieldProperty(IThesaurusTermWidget['thesaurus_name'])
    extract_name = FieldProperty(IThesaurusTermWidget['extract_name'])


@adapter_config(required=(IThesaurusTermField, IFormLayer),
                provides=IFieldWidget)
def ThesaurusTermFieldWidget(field, request):  # pylint: disable=invalid-name
    """Thesaurus term field widget factory"""
    return FieldWidget(field, ThesaurusTermWidget(request))


#
# Terms list widget
#

@adapter_config(required=(IThesaurusTermsListField, IThesaurusTermsListWidget),
                provides=IDataConverter)
class ThesaurusTermsListDataConverter(SequenceDataConverter):
    """Thesaurus terms list data converter"""

    def to_widget_value(self, value):
        # Widget expects a list of thesaurus terms labels or captions
        if value is self.field.missing_value:
            return []
        return [term.label for term in value]

    def to_field_value(self, value):
        # Field expects a list of thesaurus terms
        if not value:
            return self.field.missing_value
        thesaurus = self.widget.thesaurus
        if thesaurus is None:
            return None
        if isinstance(value, str):
            value = value.split('|')
        for idx, val in enumerate(value):
            if val == self.widget.no_value_token:
                continue
            match = SYNONYM.match(val)
            if match:
                value[idx] = match.groups()[0]
        terms = thesaurus.terms
        return [
            terms.get(term)
            for term in value
            if term != self.widget.no_value_token
        ]


@implementer_only(IThesaurusTermsListWidget)
class ThesaurusTermsListWidget(ThesaurusWidgetMixin, SelectWidget):
    """Thesaurus terms list widget"""

    ajax_url = '/api/thesaurus/terms'
    placeholder = _("No selected term")

    multiple = 'multiple'

    _thesaurus_name = FieldProperty(IThesaurusTermsListWidget['thesaurus_name'])
    extract_name = FieldProperty(IThesaurusTermsListWidget['extract_name'])


@adapter_config(required=(IThesaurusTermsListField, IFormLayer),
                provides=IFieldWidget)
def ThesaurusTermsListFieldWidget(field, request):  # pylint: disable=invalid-name
    """Thesaurus terms list field widget factory"""
    return FieldWidget(field, ThesaurusTermsListWidget(request))


@implementer_only(IThesaurusTermsTreeWidget)
class ThesaurusTermsTreeWidget(ThesaurusTermsListWidget):
    """Thesaurus terms tree widget"""

    @property
    def top_terms(self):
        """Thesaurus top terms getter"""
        thesaurus = self.thesaurus
        if thesaurus is None:
            return ()
        return sorted(thesaurus.get_top_terms(extract=self.extract_name),
                      key=lambda x: (x.order or 999, x.label))

    def get_subterms(self, term):
        """Get sub-terms from given term"""
        for subterm in term.specifics:
            if (not self.extract_name) or (self.extract_name in subterm.extracts):
                if subterm.status != STATUS_ARCHIVED:
                    yield subterm
                    yield from self.get_subterms(subterm)


def ThesaurusTermsTreeFieldWidget(field, request):  # pylint: disable=invalid-name
    """Thesaurus terms tree field widget factory"""
    return FieldWidget(field, ThesaurusTermsTreeWidget(request))



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.loader.skos module

This module is used to import or export thesaurus in SKOS/RDF format.
"""

import chardet
from lxml import etree

from pyams_thesaurus.interfaces.loader import IThesaurusExporter, IThesaurusLoader
from pyams_thesaurus.loader import BaseThesaurusExporter, BaseThesaurusLoader, \
    ThesaurusLoaderDescription, ThesaurusLoaderTerm, XMLThesaurusExporterHandler, \
    XMLThesaurusLoaderHandler
from pyams_utils.registry import utility_config


__docformat__ = 'restructuredtext'


# XML namespaces

XML = '{http://www.w3.org/XML/1998/namespace}'
RDF = '{http://www.w3.org/1999/02/22-rdf-syntax-ns#}'
RDFS = '{http://www.w3.org/2000/01/rdf-schema#}'
DC = '{http://purl.org/dc/elements/1.1/}'
DCT = '{http://purl.org/dc/terms/}'
MAP = '{http://www.w3c.rl.ac.uk/2003/11/21-skos-mapping#}'
SKOS = '{http://www.w3.org/2004/02/skos/core#}'


class SKOSThesaurusLoaderHandler(XMLThesaurusLoaderHandler):
    """SKOS format thesaurus handler"""

    def read(self, data, configuration=None):  # pylint: disable=too-many-branches,too-many-statements,too-many-locals
        """Thesaurus loader reader"""
        terms = {}
        if configuration is None:
            configuration = self.configuration
        encoding = None
        if configuration and configuration.encoding:
            encoding = configuration.encoding
        if (not encoding) and isinstance(data, str):
            encoding = chardet.detect(data[:1000]).get('encoding', 'utf-8')
        parser = etree.XMLParser(ns_clean=True, recover=True, encoding=encoding)
        xml = etree.parse(data, parser=parser)
        root = xml.getroot()
        # check thesaurus scheme
        description = ThesaurusLoaderDescription()
        scheme = root.find('.//' + SKOS + 'ConceptScheme')
        if scheme is not None:
            for element in scheme.getchildren():
                if element.tag == DC + 'title':
                    description.title = str(element.text)
                elif element.tag == DC + 'creator':
                    description.creator = str(element.text)
                elif element.tag == DC + 'subject':
                    description.subject = str(element.text)
                elif element.tag == DC + 'description':
                    description.description = str(element.text)
                elif element.tag == DC + 'publisher':
                    description.publisher = str(element.text)
                elif element.tag == DC + 'date':
                    description.created = str(element.text)
                elif element.tag == DC + 'language':
                    description.language = element.text
        if configuration and not description.language:
            description.language = configuration.language
        # check thesaurus terms
        for concept in root.findall(SKOS + 'Concept'):
            key = concept.attrib.get(RDF + 'about')
            label = None
            alt = None
            definition = None
            note = None
            generic = None
            specifics = []
            associations = []
            usage = None
            used_for = []
            created = None
            modified = None
            for element in concept.getchildren():
                if element.tag == SKOS + 'prefLabel':
                    label = str(element.text)
                elif element.tag == SKOS + 'altLabel':
                    term = element.attrib.get(RDF + 'resource')
                    if term is not None:
                        # import synonyms ?
                        if not configuration.import_synonyms:
                            continue
                        # link to another synonym resource
                        used_for.append(term)
                        if term not in terms:
                            # initialize synonym with usage field
                            terms[term] = ThesaurusLoaderTerm(term, alt='', definition='',
                                                              note='', generic='',
                                                              specifics='', associations='',
                                                              usage=key, used_for=[])
                        else:
                            terms[term].usage = key
                    else:
                        # just an alternate label
                        alt = str(element.text)
                elif element.tag == SKOS + 'definition':
                    definition = str(element.text)
                elif element.tag in (SKOS + 'note', SKOS + 'scopeNote'):
                    note = str(element.text)
                elif element.tag == SKOS + 'related':
                    associations.append(str(element.attrib[RDF + 'resource']))
                elif element.tag == SKOS + 'broader':
                    generic = str(element.attrib[RDF + 'resource'])
                elif element.tag == SKOS + 'narrower':
                    specifics.append(str(element.attrib[RDF + 'resource']))
                elif element.tag == DCT + 'created':
                    created = element.text
            if key not in terms:
                terms[key] = ThesaurusLoaderTerm(label, alt, definition, note,
                                                 generic, specifics, associations,
                                                 usage, used_for, created, modified)
            else:
                # update an already initialized synonym
                term = terms[key]
                term.label = label
                term.alt = alt
                term.definition = definition
                term.note = note
                term.generic = generic
                term.specifics = specifics
                term.associations = associations
                term.usage = usage
                term.used_for = used_for
                term.created = created
                term.modified = modified
        return description, terms


@utility_config(name='SKOS (RDF)',
                provides=IThesaurusLoader)
class SKOSThesaurusLoader(BaseThesaurusLoader):
    """SKOS format thesaurus loader"""

    handler = SKOSThesaurusLoaderHandler


class SKOSThesaurusExporterHandler(XMLThesaurusExporterHandler):
    """SKOS/RDF format thesaurus export handler"""

    def _write(self, thesaurus, configuration=None):  # pylint: disable=too-many-branches
        thesaurus_url = 'Thesaurus/{}'.format(thesaurus.__name__)
        nsmap = {
            'rdf': RDF[1:-1],
            'rdfs': RDFS[1:-1],
            'dc': DC[1:-1],
            'dct': DCT[1:-1],
            'map': MAP[1:-1],
            'skos': SKOS[1:-1]
        }
        xml = etree.Element(RDF + 'RDF', nsmap=nsmap)
        doc = etree.ElementTree(xml)
        cs = etree.SubElement(xml, SKOS + 'ConceptScheme')  # pylint: disable=invalid-name
        cs.attrib[RDF + 'about'] = thesaurus_url
        etree.SubElement(cs, DC + 'title').text = thesaurus.title
        etree.SubElement(cs, DC + 'creator').text = thesaurus.creator
        etree.SubElement(cs, DC + 'subject').text = thesaurus.subject
        if thesaurus.description:
            etree.SubElement(cs, DC + 'description').text = etree.CDATA(thesaurus.description)
        etree.SubElement(cs, DC + 'publisher').text = thesaurus.publisher
        if thesaurus.created:
            etree.SubElement(cs, DC + 'date').text = thesaurus.created.strftime('%Y-%m-%d')
        etree.SubElement(cs, DC + 'language').text = thesaurus.language
        extract = configuration.extract if configuration else None
        for term in thesaurus.top_terms:
            if extract and (extract not in (term.extracts or ())):
                continue
            etree.SubElement(cs, SKOS + 'hasTopConcept').attrib[RDF + 'resource'] = \
                term.short_url()
        for term in thesaurus.terms.values():
            if extract and (extract not in (term.extracts or ())):
                continue
            concept = etree.SubElement(xml, SKOS + 'Concept')
            concept.attrib[RDF + 'about'] = term.short_url()
            sub = etree.SubElement(concept, SKOS + 'prefLabel')
            sub.attrib[XML + 'lang'] = thesaurus.language
            sub.text = term.label
            etree.SubElement(concept, SKOS + 'inScheme').attrib[RDF + 'resource'] = thesaurus_url
            if term.definition:
                sub = etree.SubElement(concept, SKOS + 'definition')
                sub.attrib[XML + 'lang'] = thesaurus.language
                sub.text = term.definition
            if term.note:
                sub = etree.SubElement(concept, SKOS + 'scopeNote')
                sub.attrib[XML + 'lang'] = thesaurus.language
                sub.text = etree.CDATA(term.note)
            for subterm in term.associations:
                if extract and (extract not in (subterm.extracts or ())):
                    continue
                etree.SubElement(concept, SKOS + 'related').attrib[RDF + 'resource'] = \
                    subterm.short_url()
            if term.generic:
                etree.SubElement(concept, SKOS + 'broader').attrib[RDF + 'resource'] = \
                    term.generic.short_url()
            for subterm in term.used_for:
                if extract and (extract not in (subterm.extracts or ())):
                    continue
                etree.SubElement(concept, SKOS + 'altLabel').attrib[RDF + 'resource'] = \
                    subterm.short_url()
            for subterm in term.specifics:
                if extract and (extract not in (subterm.extracts or ())):
                    continue
                etree.SubElement(concept, SKOS + 'narrower').attrib[RDF + 'resource'] = \
                    subterm.short_url()
            if term.created:
                etree.SubElement(concept, DCT + 'created').text = \
                    term.created.strftime('%Y-%m-%d %H:%M:%S')
        return doc


@utility_config(name='SKOS (RDF)',
                provides=IThesaurusExporter)
class SKOSThesaurusExporter(BaseThesaurusExporter):
    """SKOS/RDF format thesaurus exporter"""

    handler = SKOSThesaurusExporterHandler



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.zmi.tests module

A test module for several PyAMS_thesaurus widgets
"""

from persistent import Persistent
from zope.interface import Interface
from zope.schema.fieldproperty import FieldProperty

from pyams_form.ajax import ajax_form_config
from pyams_form.field import Fields
from pyams_layer.interfaces import IPyAMSLayer
from pyams_security.interfaces.base import MANAGE_SYSTEM_PERMISSION
from pyams_site.interfaces import ISiteRoot
from pyams_thesaurus.interfaces.thesaurus import IThesaurus
from pyams_thesaurus.schema import ThesaurusTermsListField
from pyams_thesaurus.zmi.widget import ThesaurusTermsTreeFieldWidget
from pyams_utils.adapter import adapter_config, get_annotation_adapter
from pyams_utils.factory import factory_config
from pyams_utils.registry import query_utility
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminEditForm
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.viewlet import IPropertiesMenu
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _  # pylint: disable=ungrouped-imports


class IThesaurusTest(Interface):
    """Thesaurus test interface"""

    field1 = ThesaurusTermsListField(title=_("Thesaurus terms"),
                                     required=False)


@factory_config(provided=IThesaurusTest)
class ThesaurusTest(Persistent):
    """Thesaurus test persistent class"""

    field1 = FieldProperty(IThesaurusTest['field1'])


PYAMS_THESAURUS_TEST_KEY = 'pyams_thesaurus.test'


@adapter_config(required=ISiteRoot,
                provides=IThesaurusTest)
def site_root_thesaurus_test_adapter(context):
    """Site root thesaurus test adapter"""
    return get_annotation_adapter(context, PYAMS_THESAURUS_TEST_KEY, IThesaurusTest)


@viewlet_config(name='thesaurus-test.menu',
                context=ISiteRoot, layer=IAdminLayer,
                manager=IPropertiesMenu, weight=900,
                permission=MANAGE_SYSTEM_PERMISSION)
class ThesaurusTestMenu(NavigationMenuItem):
    """Thesaurus test menu"""

    label = _("Test thesaurus widgets")
    href = '#thesaurus-test.html'


@ajax_form_config(name='thesaurus-test.html',
                  context=ISiteRoot, layer=IPyAMSLayer,
                  permission=MANAGE_SYSTEM_PERMISSION)
class ThesaurusTestForm(AdminEditForm):
    """Thesaurus test form"""

    title = _("Thesaurus test form")
    legend = _("Thesaurus widget test")

    fields = Fields(IThesaurusTest)
    fields['field1'].widget_factory = ThesaurusTermsTreeFieldWidget

    label_css_class = 'col-md-2'
    input_css_class = 'col-md-10'

    def get_content(self):
        """Form content getter"""
        return IThesaurusTest(self.context)

    def update_widgets(self, prefix=None):
        super().update_widgets(prefix)
        field1 = self.widgets.get('field1')
        if field1 is not None:
            field1.thesaurus_name = 'Tags'


@adapter_config(name='field1',
                required=(IThesaurusTest, IAdminLayer, ThesaurusTestForm),
                provides=IThesaurus)
def thesaurus_test_thesaurus_adapter(context, request, view):  # pylint: disable=unused-argument
    """Thesaurus test thesaurus adapter"""
    return query_utility(IThesaurus, name='Tags')



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.generations main module

This module provides site generation utility to automatically create
a thesaurus container on upgrade.
"""

from pyams_site.generations import check_required_utilities
from pyams_site.interfaces import ISiteGenerations
from pyams_thesaurus.interfaces.thesaurus import IThesaurusManager, THESAURUS_MANAGER_NAME
from pyams_utils.registry import utility_config


__docformat__ = 'restructuredtext'


REQUIRED_UTILITIES = ((IThesaurusManager, '', None, THESAURUS_MANAGER_NAME),)


@utility_config(name='PyAMS thesaurus', provides=ISiteGenerations)
class ThesaurusGenerationsChecker:
    """Thesaurus generations checker"""

    order = 100
    generation = 1

    def evolve(self, site, current=None):  # pylint: disable=unused-argument,no-self-use
        """Check for required utilities"""
        check_required_utilities(site, REQUIRED_UTILITIES)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.extension.html.interfaces module

This module defines interface of an HTML term extension, which allows to provide
a rich text content on any term.
"""

from zope.annotation import IAttributeAnnotatable
from zope.interface import Interface

from pyams_i18n.schema import I18nHTMLField


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _


THESAURUS_TERM_HTML_INFO_KEY = 'pyams_thesaurus.extension.html'


class IThesaurusTermHTMLInfo(Interface):
    """Thesaurus term HTML info"""

    description = I18nHTMLField(title=_("HTML content"),
                                required=False)


class IThesaurusTermHTMLTarget(IAttributeAnnotatable):
    """Thesaurus term HTML description target"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.extension.html module

This module defines thesaurus term HTML extension.
"""

from persistent import Persistent
from zope.container.contained import Contained
from zope.schema.fieldproperty import FieldProperty

from pyams_thesaurus.extension.html.interfaces import IThesaurusTermHTMLInfo, \
    IThesaurusTermHTMLTarget, THESAURUS_TERM_HTML_INFO_KEY
from pyams_thesaurus.interfaces.extension import IThesaurusTermExtension
from pyams_utils.adapter import adapter_config, get_annotation_adapter
from pyams_utils.factory import factory_config
from pyams_utils.registry import utility_config


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _  # pylint: disable=ungrouped-imports


@factory_config(provided=IThesaurusTermHTMLInfo)
class ThesaurusTermHTMLInfo(Persistent, Contained):
    """Thesaurus term HTML description"""

    description = FieldProperty(IThesaurusTermHTMLInfo['description'])


@adapter_config(required=IThesaurusTermHTMLTarget,
                provides=IThesaurusTermHTMLInfo)
def thesaurus_term_html_description_factory(context):
    """Thesaurus term HTML description factory"""
    return get_annotation_adapter(context, THESAURUS_TERM_HTML_INFO_KEY, IThesaurusTermHTMLInfo)


@utility_config(name='html', provides=IThesaurusTermExtension)
class HTMLThesaurusExtension:
    """HTML description thesaurus extension"""

    label = _("Rich text description")
    weight = 10

    target_interface = IThesaurusTermHTMLTarget
    target_view = 'html-description.html'

    icon_css_class = 'fas fa-font'



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.loader.superdoc module

This module is used to import or export thesaurus in Superdoc format.
"""

from datetime import datetime

import chardet
from lxml import etree
from zope.intid import IIntIds

from pyams_thesaurus.interfaces.loader import IThesaurusExporter, IThesaurusLoader
from pyams_thesaurus.loader import BaseThesaurusExporter, BaseThesaurusLoader, \
    ThesaurusLoaderDescription, ThesaurusLoaderTerm, XMLThesaurusExporterHandler, \
    XMLThesaurusLoaderHandler
from pyams_utils.registry import get_utility, utility_config


__docformat__ = 'restructuredtext'


# namespaces definitions

INM = "{http://www.inmagic.com/webpublisher/query}"


class SuperdocThesaurusLoaderHandler(XMLThesaurusLoaderHandler):
    """SuperDoc format thesaurus load handler"""

    def read(self, data, configuration=None):  # pylint: disable=too-many-locals,too-many-branches,too-many-statements,too-many-nested-blocks
        """Thesaurus file read"""
        terms = {}
        if configuration is None:
            configuration = self.configuration
        encoding = None
        if configuration and configuration.encoding:
            encoding = configuration.encoding
        if (not encoding) and isinstance(data, str):
            encoding = chardet.detect(data[:1000]).get('encoding', 'utf-8')
        parser = etree.XMLParser(ns_clean=True, recover=True, encoding=encoding)
        xml = etree.parse(data, parser=parser)
        root = xml.getroot()
        # get thesaurus description
        description = ThesaurusLoaderDescription()
        if configuration:
            description.language = configuration.language
        # check thesaurus terms
        for records in root.findall(INM + 'Recordset'):  # pylint: disable=too-many-nested-blocks
            for record in records.findall(INM + 'Record'):
                key = None
                label = None
                alt = None
                definition = None
                note = None
                generic = None
                specifics = []
                associations = []
                usage = None
                used_for = []
                created = None
                modified = None
                for element in record.getchildren():
                    if element.text:
                        if element.tag == INM + 'Terme':
                            key = label = str(element.text)
                        elif element.tag == INM + 'NA':
                            definition = str(element.text)
                        elif element.tag == INM + 'TS':
                            specifics.append(str(element.text))
                        elif element.tag == INM + 'TG':
                            generic = str(element.text)
                        elif element.tag == INM + 'TA':
                            associations.append(str(element.text))
                        elif element.tag == INM + 'EM':
                            if configuration and configuration.import_synonyms:
                                usage = str(element.text)
                        elif element.tag == INM + 'EP':
                            if configuration and configuration.import_synonyms:
                                used_for.append(str(element.text))
                        elif element.tag == INM + 'Notes':
                            note = str(element.text)
                        elif element.tag == INM + 'DateCreation':
                            created = datetime.strptime(element.text, '%d/%m/%Y')
                        elif element.tag == INM + 'DateModification':
                            modified = datetime.strptime(element.text, '%d/%m/%Y')
                if key:
                    terms[key] = ThesaurusLoaderTerm(label, alt, definition, note,
                                                     generic, specifics, associations,
                                                     usage, used_for, created, modified)
        return description, terms


@utility_config(name='SuperDoc',
                provides=IThesaurusLoader)
class SuperdocThesaurusLoader(BaseThesaurusLoader):
    """SuperDoc export format thesaurus loader"""

    handler = SuperdocThesaurusLoaderHandler


class SuperdocThesaurusExporterHandler(XMLThesaurusExporterHandler):
    """SuperDoc format thesaurus export handler"""

    def _write(self, thesaurus, configuration=None):  # pylint: disable=too-many-branches,too-many-statements
        intids = get_utility(IIntIds)
        xml = etree.Element('Results', nsmap={None: INM[1:-1]},
                            productTitle='PyAMS Thesaurus Manager',
                            productVersion='1.0')
        doc = etree.ElementTree(xml)
        extract = configuration.extract if configuration else None
        if extract:
            terms = [
                term for term in thesaurus.terms.values()
                if extract in (term.extracts or set())
            ]
        else:
            terms = thesaurus.terms
        rs = etree.SubElement(xml, 'Recordset', setCount=str(len(terms)))  # pylint: disable=invalid-name
        for index, term in enumerate(thesaurus.terms.values()):
            if extract and (extract not in (term.extracts or set())):
                continue
            rec = etree.SubElement(rs, 'Record', setEntry=str(index))
            etree.SubElement(rec, 'ID').text = str(intids.queryId(term))
            etree.SubElement(rec, 'Terme').text = term.label
            etree.SubElement(rec, 'NA').text = term.note
            added_subterms = False
            if term.specifics:
                for subterm in term.specifics:
                    if extract and (extract not in (subterm.extracts or ())):
                        continue
                    etree.SubElement(rec, 'TS').text = subterm.label
                    added_subterms = True
            if not added_subterms:
                etree.SubElement(rec, 'TS')
            sub = etree.SubElement(rec, 'TG')
            if term.generic:
                sub.text = term.generic.label
            added_subterms = False
            if term.associations:
                for subterm in term.associations:
                    if extract and (extract not in (subterm.extracts or ())):
                        continue
                    etree.SubElement(rec, 'TA').text = subterm.label
                    added_subterms = True
            if not added_subterms:
                etree.SubElement(rec, 'TA')
            sub = etree.SubElement(rec, 'EM')
            if term.usage:
                sub.text = term.usage.label
            added_subterms = False
            if term.used_for:
                for subterm in term.used_for:
                    if extract and (extract not in (subterm.extracts or ())):
                        continue
                    etree.SubElement(rec, 'EP').text = subterm.label
                    added_subterms = True
            if not added_subterms:
                etree.SubElement(rec, 'EP')
            etree.SubElement(rec, 'Notes').text = term.definition
            etree.SubElement(rec, 'Status').text = term.status
            etree.SubElement(rec, 'DateCreation').text = term.created.strftime('%d/%m/%Y') \
                if term.created else ''
            etree.SubElement(rec, 'DateModification').text = term.modified.strftime('%d/%m/%Y') \
                if term.modified else ''
            etree.SubElement(rec, 'Niveau').text = str(term.level)
            etree.SubElement(rec, 'MicroThes').text = 'oui' if term.micro_thesaurus else 'NON'
            etree.SubElement(rec, 'Terme0').text = term.label \
                if term.parent is None else term.parent.label
        return doc


@utility_config(name='SuperDoc',
                provides=IThesaurusExporter)
class SuperdocThesaurusExporter(BaseThesaurusExporter):
    """SuperDoc format thesaurus exporter"""

    handler = SuperdocThesaurusExporterHandler



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.thesaurus module

This module define thesaurus utility class.
"""

import re

from hypatia.catalog import Catalog
from hypatia.query import Contains, Eq
from hypatia.text import Lexicon
from hypatia.text.parsetree import QueryError
from persistent import Persistent
from pyramid.events import subscriber
from transaction.interfaces import ITransactionManager
from zope.componentvocabulary.vocabulary import UtilityVocabulary
from zope.container.btree import BTreeContainer
from zope.container.contained import Contained
from zope.interface import implementer
from zope.lifecycleevent import IObjectAddedEvent, IObjectRemovedEvent
from zope.location import locate
from zope.location.interfaces import ISublocations
from zope.schema.fieldproperty import FieldProperty
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary
from zope.traversing.interfaces import ITraversable

from pyams_catalog.index import FieldIndexWithInterface, TextIndexWithInterface
from pyams_catalog.nltk import NltkStemmedTextProcessor
from pyams_catalog.query import CatalogResultSet, or_
from pyams_catalog.utils import index_object, unindex_object
from pyams_security.interfaces import IDefaultProtectionPolicy, IRolesPolicy, \
    IViewContextPermissionChecker
from pyams_security.property import RolePrincipalsFieldProperty
from pyams_security.security import ProtectedObjectMixin, ProtectedObjectRoles
from pyams_site.interfaces import ISiteRoot
from pyams_thesaurus.interfaces import ADMIN_THESAURUS_PERMISSION, \
    THESAURUS_EXTRACTS_VOCABULARY, THESAURUS_NAMES_VOCABULARY, ThesaurusTermDeleteException
from pyams_thesaurus.interfaces.extension import IThesaurusTermExtension
from pyams_thesaurus.interfaces.loader import IThesaurusLoader
from pyams_thesaurus.interfaces.term import IThesaurusLoaderTerm, IThesaurusTerm, \
    IThesaurusTermsContainer
from pyams_thesaurus.interfaces.thesaurus import IThesaurus, IThesaurusContextManager, \
    IThesaurusExtract, IThesaurusExtractRoles, IThesaurusExtracts, IThesaurusRoles
from pyams_utils.adapter import ContextAdapter, adapter_config, get_annotation_adapter
from pyams_utils.factory import factory_config
from pyams_utils.interfaces.tree import INode, ITree
from pyams_utils.registry import query_utility
from pyams_utils.request import check_request
from pyams_utils.traversing import get_parent
from pyams_utils.unicode import translate_string
from pyams_utils.url import absolute_url
from pyams_utils.vocabulary import LocalUtilitiesVocabulary, vocabulary_config
from pyams_zmi.interfaces import IObjectLabel


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _


CUSTOM_SEARCH = re.compile(r'\*|\"|\sand\s|\sor\s|\snot\s|\(|\)', re.IGNORECASE)


@implementer(IThesaurusTermsContainer)
class ThesaurusTermsContainer(BTreeContainer):
    """Thesaurus terms container"""

    def clear(self):
        """Terms container clear"""
        self._SampleContainer__data.clear()
        self._BTreeContainer__len.set(0)  # pylint: disable=no-member


@factory_config(IThesaurus)  # pylint: disable=too-many-instance-attributes
@implementer(ISublocations, IDefaultProtectionPolicy)
class Thesaurus(ProtectedObjectMixin, Persistent, Contained):
    """Thesaurus persistent class"""

    name = FieldProperty(IThesaurus['name'])
    title = FieldProperty(IThesaurus['title'])
    subject = FieldProperty(IThesaurus['subject'])
    description = FieldProperty(IThesaurus['description'])
    language = FieldProperty(IThesaurus['language'])
    creator = FieldProperty(IThesaurus['creator'])
    publisher = FieldProperty(IThesaurus['publisher'])
    created = FieldProperty(IThesaurus['created'])

    terms = None
    _top_terms = FieldProperty(IThesaurus['top_terms'])

    catalog = FieldProperty(IThesaurus['catalog'])

    def __init__(self, name=None, description=None, terms=None):
        if name:
            self.name = name
        if description:
            self.title = description.title
            self.subject = description.subject
            self.description = description.description
            self.language = description.language
            self.creator = description.creator
            self.publisher = description.publisher
            self.created = description.created
        if not IThesaurusTermsContainer.providedBy(terms):  # pylint: disable=no-value-for-parameter
            terms = ThesaurusTermsContainer()
        self.terms = terms
        locate(terms, self, '++terms++')
        self.reset_terms_parent()
        self.reset_top_terms()

    def sublocations(self):
        """Sub-locations getter"""
        yield from (self.terms, self.catalog)

    @property
    def top_terms(self):
        """Top terms getter"""
        return self._top_terms

    @top_terms.setter
    def top_terms(self, value):
        """Top terms setter"""
        self._top_terms = [term for term in value or () if term.usage is None]

    def init_catalog(self):
        """Inner catalog initialization"""
        # Init catalog
        catalog = self.catalog = Catalog()
        locate(catalog, self)
        # Init fulltext search catalog
        catalog['fulltext'] = TextIndexWithInterface(IThesaurusTerm, 'label',
                                                     language=self.language)
        # Init stemmed search catalog
        catalog['stemmed'] = TextIndexWithInterface(IThesaurusTerm, 'label',
                                                    lexicon=Lexicon(NltkStemmedTextProcessor(
                                                        language=self.language)))
        # Init value search catalog
        catalog['value'] = FieldIndexWithInterface(IThesaurusTerm, 'base_label')
        for idx, term in enumerate(self.terms.values()):
            index_object(term, catalog)
            if not idx % 100:
                try:
                    ITransactionManager(catalog).savepoint()
                except TypeError:
                    # Can fail if thesaurus is not stored yet...
                    pass

    def load(self, configuration):
        """Load terms from given configuration"""
        loader = query_utility(IThesaurusLoader, name=configuration.format)
        if loader is not None:
            result = loader.load(configuration.data, configuration)
            self.merge(configuration, result)

    def merge(self, configuration, thesaurus=None):
        """Merge existing thesaurus terms with imported configuration"""
        if thesaurus is None:
            loader = query_utility(IThesaurusLoader, name=configuration.format)
            if loader is not None:
                thesaurus = loader.load(configuration.data)
        if thesaurus is not None:
            # define or merge items from given thesaurus
            terms = self.terms
            for idx, (key, term) in enumerate(thesaurus.terms.items()):
                # check for term conflict
                suffix = getattr(configuration, 'conflict_suffix', None)
                if suffix:
                    suffixed_key = key + ' ' + suffix
                    if suffixed_key in terms:
                        key = suffixed_key
                    elif key in terms:
                        term.label = key
                if key in terms:
                    terms[key].merge(term, configuration)
                elif not IThesaurusLoaderTerm.providedBy(term):  # pylint: disable=no-value-for-parameter
                    terms[key] = term
                if not idx % 100:
                    try:
                        ITransactionManager(self).savepoint()
                    except TypeError:
                        # Can fail if thesaurus is not stored yet...
                        pass
        self.reset_terms_parent()
        self.reset_top_terms()

    def reset_terms_parent(self):
        """Reset parent from thesaurus terms"""
        for idx, term in enumerate(self.terms.values()):
            # reset generic/specifics attributes
            generic = term.generic
            if (generic is not None) and (term not in generic.specifics):
                generic.specifics = generic.specifics + [term, ]
            # reset term's first level parent
            parent = term
            while parent.generic is not None:
                parent = parent.generic
            term.parent = parent
            if not idx % 100:
                try:
                    ITransactionManager(self).savepoint()
                except TypeError:
                    # Can fail if thesaurus is not stored yet...
                    pass

    def reset_top_terms(self):
        """Reset thesaurus top terms"""
        self.top_terms = [
            term for term in self.terms.values()
            if (not term.generic) and (not term.usage)
        ]

    def get_top_terms(self, extract=None):
        """Get thesaurus top terms"""
        if not extract:
            return self.top_terms
        return [term for term in self.top_terms if extract in term.extracts]

    def remove_term(self, term):
        """Remove specified term from thesaurus"""
        if term.label not in self.terms:
            raise ThesaurusTermDeleteException(_("Term not found"))
        if not term.is_deletable():
            raise ThesaurusTermDeleteException(_("Term can't be deleted"))
        unindex_object(term, self.catalog)
        # check generic
        generic = term.generic
        if generic is not None:
            specifics = generic.specifics
            specifics.remove(term)
            generic.specifics = specifics
        # check top terms
        terms = self.top_terms
        if term in terms:
            terms.remove(term)
            self.top_terms = terms
        # delete term
        del self.terms[term.label]

    def clear(self):
        """Remove all terms from thesaurus"""
        self.terms.clear()
        self.catalog.reset()
        self.top_terms = []

    def find_terms(self, query=None, extract=None, glob='end', limit=None,  # pylint: disable=too-many-arguments
                   exact=False, exact_only=False, stemmed=False):
        """Find terms matching given criteria"""
        assert exact or (not exact_only)
        terms = []
        if exact:
            query_text = translate_string(query, escape_slashes=True,
                                          force_lower=True, spaces=' ')
            terms = list(CatalogResultSet(Eq(self.catalog['value'], query_text)))
        if not exact_only:
            search = None
            # check stemmed index
            if stemmed and not re.search(r'\*', query):
                search = Contains(self.catalog['stemmed'],
                                  ' and '.join(m
                                               for m in query.split()
                                               if len(m) > 2))
            # check basic index
            start = ''
            end = ''
            if CUSTOM_SEARCH.search(query):
                query_text = query
            else:
                if glob in ('start', 'both'):
                    # Starting glob is not supported!!
                    start = ''
                if glob in ('end', 'both'):
                    end = '*'
                query_text = ' and '.join((
                    '{start}{mid}{end}'.format(start=start, mid=m, end=end)
                    for m in query.split()
                    if len(m) > 2
                ))
            search = or_(search, Contains(self.catalog['fulltext'], query_text))
            try:
                terms += sorted(CatalogResultSet(search.execute()),
                                key=lambda x: x.label)
            except QueryError:
                pass
        if extract:
            terms = filter(lambda term: extract in term.extracts, terms)
        if limit:
            terms = terms[:limit]
        return terms

    def get_nodes(self, term, result, subnodes=None, request=None):
        """Get nodes matching given parent term"""
        if request is None:
            request = check_request()
        registry = request.registry
        translate = request.localizer.translate
        extracts = IThesaurusExtracts(self)
        for child in INode(term).get_children():
            node = INode(child)
            result.append({
                'label': node.label.replace("'", "&#039;"),
                'view': absolute_url(node.context, request, 'properties.html'),
                'css_class': node.css_class,
                'extracts': [
                    {
                        'name': name,
                        'title': extract.name,
                        'color': extract.color,
                        'used': name in (node.context.extracts or ())
                    }
                    for name, extract in sorted(extracts.items(), key=lambda x: x[0])
                ],
                'extensions': [
                    {
                        'name': name,
                        'title': translate(ext.label),
                        'icon': ext.icon_css_class,
                        'view': absolute_url(node.context, request, ext.target_view),
                        'active': name in node.context.extensions
                    }
                    for name, ext in sorted(registry.getUtilitiesFor(IThesaurusTermExtension),
                                            key=lambda ext: ext[1].weight,
                                            reverse=True)
                ],
                'expand': node.has_children()
            })
            if subnodes and (node.context.label in subnodes):
                nodes = result[-1]['subnodes'] = []
                self.get_nodes(node.context, nodes, subnodes, request)

    def delete(self):
        """Delete thesaurus"""


@adapter_config(required=IThesaurus,
                provides=IObjectLabel)
def thesaurus_label(context):
    """Thesaurus label getter"""
    return context.name


@implementer(IThesaurusRoles)
class ThesaurusRoles(ProtectedObjectRoles):
    """Thesaurus roles"""

    administrators = RolePrincipalsFieldProperty(IThesaurusRoles['administrators'])
    managers = RolePrincipalsFieldProperty(IThesaurusRoles['managers'])


@adapter_config(required=IThesaurus,
                provides=IThesaurusRoles)
def thesaurus_roles_adapter(context):
    """Thesaurus roles adapter"""
    return ThesaurusRoles(context)


@adapter_config(name='thesaurus_roles',
                required=IThesaurus,
                provides=IRolesPolicy)
class ThesaurusRolesPolicy(ContextAdapter):
    """Thesaurus roles policy"""

    roles_interface = IThesaurusRoles
    weight = 20


@subscriber(IObjectAddedEvent, context_selector=IThesaurus)
def handle_added_thesaurus(event):
    """Handle added thesaurus to init inner catalog"""
    sm = get_parent(event.newParent, ISiteRoot)  # pylint: disable=invalid-name
    if sm is not None:
        thesaurus = event.object
        sm.getSiteManager().registerUtility(thesaurus, IThesaurus,
                                            name=thesaurus.name or '')
        thesaurus.init_catalog()


@subscriber(IObjectRemovedEvent, context_selector=IThesaurus)
def handle_removed_thesaurus(event):
    """Handle removed thesaurus"""
    sm = get_parent(event.oldParent, ISiteRoot)  # pylint: disable=invalid-name
    if sm is not None:
        thesaurus = event.object
        sm.getSiteManager().unregisterUtility(thesaurus, IThesaurus,
                                              name=thesaurus.name or '')


@adapter_config(required=IThesaurus,
                provides=IViewContextPermissionChecker)
class ThesaurusPermissionChecker(ContextAdapter):
    """Thesaurus edit permission checker"""

    edit_permission = ADMIN_THESAURUS_PERMISSION


@adapter_config(required=IThesaurus,
                provides=ITree)
class ThesaurusTreeAdapter(ContextAdapter):
    """Thesaurus tree adapter"""

    def get_root_nodes(self):
        """Thesaurus root nodes getter"""
        return sorted(self.context.top_terms, key=lambda x: (x.order or 999, x.label))


#
# Thesaurus extracts
#

THESAURUS_EXTRACTS_KEY = 'pyams_thesaurus.extracts'


@factory_config(IThesaurusExtracts)
class ThesaurusExtractsContainer(BTreeContainer):
    """Thesaurus extracts container"""


@adapter_config(required=IThesaurus,
                provides=IThesaurusExtracts)
def thesaurus_extracts_factory(context):
    """Thesaurus extracts factory"""
    return get_annotation_adapter(context, THESAURUS_EXTRACTS_KEY, IThesaurusExtracts,
                                  name='++extracts++')


@adapter_config(name='extracts',
                required=IThesaurus,
                provides=ITraversable)
class ThesaurusExtractsNamespace(ContextAdapter):
    """Thesaurus ++extracts++ namespace"""

    def traverse(self, name, furtherpath=None):  # pylint: disable=unused-argument
        """Thesaurus extracts traverse"""
        extracts = IThesaurusExtracts(self.context)
        if name:
            return extracts[name]
        return extracts


@factory_config(IThesaurusExtract)
@implementer(IDefaultProtectionPolicy)
class ThesaurusExtract(ProtectedObjectMixin, Persistent, Contained):
    """Thesaurus extract"""

    name = FieldProperty(IThesaurusExtract['name'])
    description = FieldProperty(IThesaurusExtract['description'])
    abbreviation = FieldProperty(IThesaurusExtract['abbreviation'])
    color = FieldProperty(IThesaurusExtract['color'])

    def add_term(self, term):
        """Add term to extract"""
        term.add_extract(self)

    def remove_term(self, term):
        """Remove term from extract"""
        term.remove_extract(self)

    def get_nodes(self, term, result, subnodes=None, request=None):
        """Extract nodes getter"""
        if request is None:
            request = check_request()
        extract_name = self.name
        for child in INode(term).get_children():
            node = INode(child)
            if extract_name in (node.context.extracts or ()):
                result.append({
                    'label': node.label.replace("'", "&#039;"),
                    'view': absolute_url(node.context, request, 'properties.html'),
                    'css_class': node.css_class,
                    'extracts': [],
                    'extensions': [],
                    'expand': node.has_children(extract_name)
                })
                if subnodes and (node.context.label in subnodes):
                    nodes = result[-1]['subnodes'] = []
                    self.get_nodes(node.context, nodes, subnodes, request)


@adapter_config(required=IThesaurusExtract,
                provides=IViewContextPermissionChecker)
class ThesaurusExtractPermissionChecker(ContextAdapter):
    """Thesaurus extract permission checker"""

    edit_permission = ADMIN_THESAURUS_PERMISSION


@implementer(IThesaurusExtractRoles)
class ThesaurusExtractRoles(ProtectedObjectRoles):
    """Thesaurus extract roles"""

    managers = RolePrincipalsFieldProperty(IThesaurusExtractRoles['managers'])


@adapter_config(required=IThesaurusExtract,
                provides=IThesaurusExtractRoles)
def thesaurus_extract_roles_adapter(context):
    """Thesaurus extract roles adapter"""
    return ThesaurusExtractRoles(context)


@adapter_config(name='thesaurus_extract_roles',
                required=IThesaurusExtract,
                provides=IRolesPolicy)
class ThesaurusExtractRolesPolicy(ContextAdapter):
    """Thesaurus extract roles policy"""

    roles_interface = IThesaurusExtractRoles
    weight = 20


@subscriber(IObjectRemovedEvent, context_selector=IThesaurusExtract)
def handle_removed_extract(event):
    """Handle removed extract"""
    thesaurus = get_parent(event.object, IThesaurus)
    name = event.object.name
    for term in thesaurus.terms.values():
        term.remove_extract(name, check=False)


#
# Vocabularies
#

@vocabulary_config(name=THESAURUS_NAMES_VOCABULARY)
class ThesaurusNamesVocabulary(LocalUtilitiesVocabulary):
    """Thesaurus names utilities vocabulary"""

    interface = IThesaurus

    @staticmethod
    def title_factory(thesaurus):
        """Thesaurus title getter"""
        return thesaurus.title


@vocabulary_config(name=THESAURUS_EXTRACTS_VOCABULARY)
class ThesaurusExtractsVocabulary(SimpleVocabulary):
    """Thesaurus extracts vocabulary"""

    def __init__(self, context=None):
        terms = []
        if context is not None:
            request = check_request()
            thesaurus_name = request.headers.get('X-Thesaurus-Name')
            if thesaurus_name is not None:
                thesaurus = query_utility(IThesaurus, name=thesaurus_name)
            else:
                thesaurus = get_parent(context, IThesaurus)
                if thesaurus is None:
                    context_manager = IThesaurusContextManager(context, None)
                    if (context_manager is not None) and context_manager.thesaurus_name:
                        thesaurus = query_utility(IThesaurus,
                                                  name=context_manager.thesaurus_name)
            if thesaurus is not None:
                extracts = IThesaurusExtracts(thesaurus)
                terms = [
                    SimpleTerm(extract.__name__, title=extract.name)
                    for extract in extracts.values()
                ]
                terms.sort(key=lambda x: x.title)
        super().__init__(terms)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.zmi.extension module

This module defines base components for extensions management.
"""
from pyams_thesaurus.interfaces import MANAGE_THESAURUS_CONTENT_PERMISSION
from pyams_thesaurus.interfaces.thesaurus import IThesaurus
from pyams_utils.traversing import get_parent
from pyams_zmi.form import AdminModalEditForm


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _  # pylint: disable=ungrouped-imports


class ThesaurusTermExtensionEditForm(AdminModalEditForm):
    """Thesaurus term extension properties edit form"""

    @property
    def title(self):
        """Title getter"""
        translate = self.request.localizer.translate
        thesaurus = get_parent(self.context, IThesaurus)
        return '<small>{}</small><br />{}'.format(
            translate(_("Thesaurus: {}")).format(thesaurus.name),
            translate(_("Term: {}")).format(self.context.label))

    _edit_permission = MANAGE_THESAURUS_CONTENT_PERMISSION



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.interfaces.extension module

This module defines thesaurus interfaces related to extensions management, which
are utilities which can be used to add additional information to thesaurus terms.
"""

from zope.interface import Attribute, Interface
from zope.schema import TextLine, URI


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _


THESAURUS_EXTENSIONS_VOCABULARY = 'pyams_thesaurus.extensions'


class IThesaurusTermExtension(Interface):
    """Thesaurus term extension interface

    An extension is a marker interface implemented by
    a term, which provides additional attributes to the term.

    Each available extension is defined as a named utility.
    """

    label = TextLine(title=_("Extension name"),
                     description=_("User name given to the extension"),
                     required=True)

    target_interface = Attribute(_("Extension marker interface"))

    target_view = URI(title=_("Extension target view name"),
                      required=True)

    icon_css_class = TextLine(title=_("Extension icon CSS class"),
                              required=True)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.interfaces.loader module

This module defines interfaces of utilities used to import or export
thesaurus contents into several formats.
"""

from zope.interface import Attribute, Interface
from zope.schema import Bool, Choice, TextLine

from pyams_file.schema import FileField
from pyams_i18n.interfaces import BASE_LANGUAGES_VOCABULARY_NAME
from pyams_thesaurus.interfaces import THESAURUS_EXTRACTS_VOCABULARY
from pyams_thesaurus.schema import ValidatedChoice
from pyams_utils.schema import EncodingField


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _  # pylint: disable=ungrouped-imports


THESAURUS_LOADERS_VOCABULARY = 'pyams_thesaurus.loaders'


class IThesaurusLoaderConfiguration(Interface):
    """Thesaurus loader configuration interface"""

    name = TextLine(title=_("Thesaurus name"),
                    description=_("Name of the registered thesaurus"),
                    required=True)

    data = FileField(title=_("Input data"),
                     description=_("Input file containing thesaurus data"),
                     required=True)

    format = Choice(title=_("File format"),
                    description=_("This list contains available thesauri loaders"),
                    required=True,
                    vocabulary=THESAURUS_LOADERS_VOCABULARY)

    import_synonyms = Bool(title=_("Import synonyms?"),
                           description=_("If 'No', synonyms will not be imported into loaded "
                                         "thesaurus"),
                           required=True,
                           default=True)

    language = Choice(title=_("Content language"),
                      description=_("Select file language, for formats which don't provide "
                                    "it internally"),
                      required=False,
                      vocabulary=BASE_LANGUAGES_VOCABULARY_NAME)

    encoding = EncodingField(title=_("File encoding"),
                             description=_("Select file encoding, for formats which don't "
                                           "provide it internally"),
                             required=False,
                             default='utf-8')


class IThesaurusUpdaterConfiguration(IThesaurusLoaderConfiguration):
    """Thesaurus updater configuration interface"""

    clear = Bool(title=_("Clear before merge ?"),
                 description=_("If 'yes', thesaurus will be cleared before re-importing file "
                               "contents"),
                 required=True,
                 default=False)

    conflict_suffix = TextLine(title=_("Auto-added conflict suffix"),
                               description=_("If you want to prevent imports conflicts, you "
                                             "can provide a suffix which will be added "
                                             "automatically to conflicting terms"),
                               required=False)


class IThesaurusLoaderHandler(Interface):
    """Thesaurus loader handler configuration"""

    configuration = Attribute("Current handler configuration")

    def read(self, data, configuration=None):
        """Extract terms from given data"""


class IThesaurusLoader(Interface):
    """Thesaurus loader interface"""

    handler = Attribute("Thesaurus handler class")

    def load(self, data, configuration=None):
        """Load thesaurus from data for the given loader configuration"""


THESAURUS_EXPORTERS_VOCABULARY = 'pyams_thesaurus.exporters'


class IThesaurusExporterConfiguration(Interface):
    """Thesaurus exporter configuration interface"""

    filename = TextLine(title=_("Export file name"),
                        description=_("Full file name, including extension"),
                        required=False)

    format = Choice(title=_("Export file format"),
                    description=_("This list contains available thesauri exporters"),
                    required=True,
                    vocabulary=THESAURUS_EXPORTERS_VOCABULARY)

    extract = ValidatedChoice(title=_("Extract to export"),
                              description=_("You can choose to export only an extract of the "
                                            "thesaurus"),
                              required=False,
                              vocabulary=THESAURUS_EXTRACTS_VOCABULARY)


class IThesaurusExporterHandler(Interface):
    """Thesaurus exporter handler configuration"""

    content_type = Attribute("Content type of export file")

    configuration = Attribute("Current handler configuration")

    def write(self, thesaurus, output, configuration=None):
        """Export terms of given thesaurus"""


class IThesaurusExporter(Interface):
    """Thesaurus exporter configuration"""

    handler = Attribute("Thesaurus handler class")

    def export(self, thesaurus, configuration=None):
        """Export thesaurus terms with the given export configuration"""
